# Generated by Django 2.0.6 on 2018-06-28 20:15

from django.db import migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('openid_connect_op', '0007_auto_20180501_0632'),
    ]

    operations = [
        migrations.AlterField(
            model_name='openidagreement',
            name='allowed_claims',
            field=jsonfield.fields.JSONField(default=[], verbose_name='List of allowed claims for this agreement'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-12-04 19:48
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openid_connect_op', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='openidclient',
            name='client_code',
            field=models.CharField(blank=True, max_length=32, null=True),
        ),
    ]



import datetime
from functools import lru_cache

import python_jwt as jwt
import jwcrypto.jwk as jwk
from django.conf import settings

from os import urandom
from jwcrypto.jws import JWS
from jwcrypto.common import base64url_encode, json_encode, json_decode
from calendar import timegm


# need to add "kid" header which the original python_jwt can not do
from openid_connect_op.models import OpenIDClient


def generate_jwt_patched(claims, priv_key=None,
                         algorithm='PS512', lifetime=None, expires=None,
                         not_before=None,
                         jti_size=16, extra_headers={}):
    """
    Generate a JSON Web Token.

    :param claims: The claims you want included in the signature.
    :type claims: dict

    :param priv_key: The private key to be used to sign the token. Note: if you pass ``None`` then the token will be returned with an empty cryptographic signature and :obj:`algorithm` will be forced to the value ``none``.
    :type priv_key: `jwcrypto.jwk.JWK <https://jwcrypto.readthedocs.io/en/latest/jwk.html>`_

    :param algorithm: The algorithm to use for generating the signature. ``RS256``, ``RS384``, ``RS512``, ``PS256``, ``PS384``, ``PS512``, ``ES256``, ``ES384``, ``ES512``, ``HS256``, ``HS384``, ``HS512`` and ``none`` are supported.
    :type algorithm: str

    :param lifetime: How long the token is valid for.
    :type lifetime: datetime.timedelta

    :param expires: When the token expires (if :obj:`lifetime` isn't specified)
    :type expires: datetime.datetime

    :param not_before: When the token is valid from. Defaults to current time (if ``None`` is passed).
    :type not_before: datetime.datetime

    :param jti_size: Size in bytes of the unique token ID to put into the token (can be used to detect replay attacks). Defaults to 16 (128 bits). Specify 0 or ``None`` to omit the JTI from the token.
    :type jti_size: int

    :rtype: unicode
    :returns: The JSON Web Token. Note this includes a header, the claims and a cryptographic signature. The following extra claims are added, per the `JWT spec <http://self-issued.info/docs/draft-ietf-oauth-json-web-token.html>`_:

    - **exp** (*IntDate*) -- The UTC expiry date and time of the token, in number of seconds from 1970-01-01T0:0:0Z UTC.
    - **iat** (*IntDate*) -- The UTC date and time at which the token was generated.
    - **nbf** (*IntDate*) -- The UTC valid-from date and time of the token.
    - **jti** (*str*) -- A unique identifier for the token.
    """
    header = {
        'typ': 'JWT',
        'alg': algorithm if priv_key else 'none'
    }
    header.update(extra_headers)

    claims = dict(claims)

    now = datetime.datetime.utcnow()

    if jti_size:
        claims['jti'] = base64url_encode(urandom(jti_size))

    claims['nbf'] = timegm((not_before or now).utctimetuple())
    claims['iat'] = timegm(now.utctimetuple())

    if lifetime:
        claims['exp'] = timegm((now + lifetime).utctimetuple())
    elif expires:
        claims['exp'] = timegm(expires.utctimetuple())

    if header['alg'] == 'none':
        signature = ''
    else:
        token = JWS(json_encode(claims))
        token.add_signature(priv_key, protected=header)
        signature = json_decode(token.serialize())['signature']

    return u'%s.%s.%s' % (
        base64url_encode(json_encode(header)),
        base64url_encode(json_encode(claims)),
        signature
    )


class JWTTools:

    @staticmethod
    def generate_jwt(payload, for_client=None, ttl=None, from_client=None):
        if for_client is None:
            sign_alg = 'RS256'
        elif for_client.client_auth_type == OpenIDClient.CLIENT_AUTH_TYPE_SECRET_JWT:
            sign_alg = 'HS256'
        else:
            sign_alg = for_client.client_registration_data.get('id_token_signed_response_alg', 'RS256')

        return JWTTools.generate_jwt_with_sign_alg(payload, sign_alg, ttl=ttl, client=from_client)

    @staticmethod
    def generate_jwt_with_sign_alg(payload, sign_alg, ttl=None, client=None):
        from openid_connect_op.models import OpenIDClient
        if not client:
            client = OpenIDClient.self_instance()
        if client.client_auth_type == client.CLIENT_AUTH_TYPE_SECRET_JWT:
            sign_key = jwk.JWK(kty="oct", use="sig", alg="HS256", k=base64url_encode(client.client_hashed_secret))
            extra_headers = {}
            alg = 'HS256'
        else:
            sign_key = client.get_key(sign_alg)
            extra_headers = {
                'kid': sign_key.key_id
            }
            alg = sign_key._params['alg']

        return generate_jwt_patched(payload,
                                    sign_key,
                                    alg,
                                    extra_headers=extra_headers,
                                    lifetime=ttl)

    @staticmethod
    def validate_jwt(token, client=None):
        from openid_connect_op.models import OpenIDClient
        if client is None:
            client = OpenIDClient.self_instance()

        if client.client_auth_type == OpenIDClient.CLIENT_AUTH_TYPE_SECRET_JWT:
            key = jwk.JWK(kty="oct", use="sig", alg="HS256", k=base64url_encode(client.client_hashed_secret))
            return jwt.verify_jwt(token, key, ['HS256'], checks_optional=True)

        header, __ = jwt.process_jwt(token)
        key = client.get_key(alg=header.get('alg', 'RS256'), kid=header.get('kid', None))
        return jwt.verify_jwt(token, key, [key._params.get('alg', 'RS256')], checks_optional=True)

    @staticmethod
    def unverified_jwt_payload(token):
        return jwt.process_jwt(token)[1]


import json

try:
    import secrets
except ImportError:
    import openid_connect_op.utils.secrets_backport as secrets

import jwcrypto.jwk as jwk
from django.conf import settings
from django.core.management import BaseCommand

from openid_connect_op.models import OpenIDClient


class Command(BaseCommand):
    def handle(self, *args, **kwargs):
        """
        Creates JWT keys at the location pointed by settings.OPENID_JWT_PRIVATE_KEY, settings.OPENID_JWT_PUBLIC_KEY
        """
        jwks = jwk.JWKSet()

        jwks['keys'].add(jwk.JWK.generate(kty='RSA', alg='RS256', size=2048, kid=secrets.token_urlsafe(32)))
        jwks['keys'].add(jwk.JWK.generate(kty='RSA', alg='RS512', size=4096, kid=secrets.token_urlsafe(32)))
        jwks['keys'].add(jwk.JWK.generate(kty='EC', crv='P-256', alg='ES256', kid=secrets.token_urlsafe(32)))
        jwks['keys'].add(jwk.JWK.generate(kty='EC', crv='P-384', alg='ES384', kid=secrets.token_urlsafe(32)))
        jwks['keys'].add(jwk.JWK.generate(kty='EC', crv='P-521', alg='ES512', kid=secrets.token_urlsafe(32)))
        jwks['keys'].add(jwk.JWK.generate(kty='oct', alg='AES', size=16*8, kid=secrets.token_urlsafe(32)))

        client = OpenIDClient.objects.get_or_create(client_id=OpenIDClient.SELF_CLIENT_ID,
                                                    defaults={
                                                        'client_auth_type': OpenIDClient.CLIENT_AUTH_TYPE_INVALID,
                                                        'client_name': 'This server'
                                                    })[0]
        client.jwks = json.dumps(json.loads(jwks.export(private_keys=True)), indent=True)
        client.save()
        # print(client.jwks)



import datetime
import json
from urllib.parse import urlencode, splitquery, parse_qs

import pytest
from django.contrib.auth.models import User
from django.core.management import call_command

from openid_connect_op.models import OpenIDClient, OpenIDToken
from openid_connect_op.utils.jwt import JWTTools
try:
    import secrets
except ImportError:
    import openid_connect_op.utils.secrets_backport as secrets

import jwcrypto.jwk as jwk


@pytest.mark.django_db
class TestTokenRequestClientSecretJWT:

    @pytest.fixture(autouse=True)
    def init_jwk(self):
        call_command('create_jwt_keys')

    @pytest.fixture
    def user(self):
        return User.objects.create(username='a')

    @pytest.fixture()
    def client_config(self):
        redirect_uri = 'http://localhost:8000/complete/test/?state=1234'

        ret = OpenIDClient.objects.create(
            client_id='test',
            redirect_uris=redirect_uri,
            client_auth_type=OpenIDClient.CLIENT_AUTH_TYPE_SECRET_JWT,
        )
        ret.set_client_secret('b')
        ret.save()
        return ret

    def test_logged_user(self, client, client_config, user, settings):
        code = self.get_authorization_code(client, client_config, user)

        token = {
            'iss': client_config.client_id,
            'sub': client_config.client_id,
            'aud': ['http://testserver/openid/token'],
        }
        jwt_token = JWTTools.generate_jwt(token, client_config, datetime.timedelta(seconds=60),
                                          from_client=client_config)

        resp = client.get('/openid/token?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'grant_type': 'authorization_code',
            'code': code,
            'client_assertion_type': 'urn:ietf:params:oauth:client-assertion-type:jwt-bearer',
            'client_assertion': jwt_token
        }))

        self.check_token_response(settings, client_config, resp)

    @staticmethod
    def check_token_response(settings, client_config, resp, sub='a'):
        assert resp.status_code == 200
        data = json.loads(resp.content.decode('utf-8'))
        assert 'access_token' in data
        assert data['token_type'] == 'Bearer'
        assert 'refresh_token' in data
        assert data['expires_in'] == settings.OPENID_DEFAULT_ACCESS_TOKEN_TTL
        assert 'id_token' in data
        database_at = OpenIDToken.objects.get(token_hash=OpenIDToken.get_token_hash(data['access_token']))
        assert database_at.user.username == 'a'
        assert database_at.client == client_config
        assert database_at.token_type == OpenIDToken.TOKEN_TYPE_ACCESS_BEARER_TOKEN
        database_rt = OpenIDToken.objects.get(token_hash=OpenIDToken.get_token_hash(data['refresh_token']))
        assert database_rt.user.username == 'a'
        assert database_rt.client == client_config
        assert database_rt.token_type == OpenIDToken.TOKEN_TYPE_REFRESH_TOKEN
        # validate id token
        header, payload = JWTTools.validate_jwt(data['id_token'])
        assert header['alg'] == 'RS256'
        assert header['typ'] == 'JWT'
        assert payload['exp'] == int(payload['exp'])
        assert payload['iat'] == int(payload['iat'])
        assert payload['aud'] == ['test']
        assert payload['sub'] == sub  # username
        assert payload['iss'] == 'http://testserver/'

    @staticmethod
    def get_authorization_code(client, client_config, user):
        client.force_login(user)
        resp = client.get('/openid/authorize?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'client_id': 'test',
            'scope': 'openid',
            'response_type': 'code'
        }))
        assert resp.status_code == 302
        redirect_server, redirect_query = splitquery(resp.url)
        assert redirect_server == 'http://localhost:8000/complete/test/'
        redirect_query = parse_qs(redirect_query)
        assert redirect_query['state'] == ['1234']
        assert 'code' in redirect_query
        code = redirect_query['code'][0]
        return code



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-01-02 18:55
from __future__ import unicode_literals

from django.db import migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('openid_connect_op', '0002_openidclient_client_code'),
    ]

    operations = [
        migrations.AddField(
            model_name='openidclient',
            name='client_registration_data',
            field=jsonfield.fields.JSONField(default={}),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-01-03 15:22
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openid_connect_op', '0005_auto_20180103_0908'),
    ]

    operations = [
        migrations.AlterField(
            model_name='openidclient',
            name='jwks',
            field=models.TextField(default='{}'),
        ),
    ]



from django.http.response import JsonResponse, HttpResponse
from django.views import View

from openid_connect_op.models import OpenIDClient
from openid_connect_op.utils.jwt import JWTTools


class JWKSView(View):
    def get(self, request, *args, **kwargs):
        client = OpenIDClient.objects.get(client_id=OpenIDClient.SELF_CLIENT_ID)
        keys = client.get_keys()
        for k in list(keys['keys']):
            if not k.has_public:
                keys['keys'].remove(k)
        return HttpResponse(keys.export(private_keys=False), content_type='application/json')


import json

import pytest
import time
from django.core.management import call_command
import jwcrypto.jwk as jwk

from openid_connect_op.utils.jwt import JWTTools


@pytest.mark.django_db
class TestJWKS:

    @pytest.fixture(autouse=True)
    def init_jwk(self):
        call_command('create_jwt_keys')

    def test_get_jwks(self, client):
        resp = client.get('/openid/jwks')
        assert resp.status_code == 200
        data = json.loads(resp.content.decode('utf-8'))
        assert 'keys' in data
        assert len(data['keys']) >= 1
        for key in data['keys']:
            jwk.JWK(**key).export_public() == key

    def test_jwks_sign(self):
        payload = {
            'a': 'b',
            'exp': time.time() + 60
        }
        for alg in ('RS256', 'RS512', 'ES256', 'ES384', 'ES512'):
            token = JWTTools.generate_jwt_with_sign_alg(payload, alg)
            print(token)
            decrypted_payload = JWTTools.validate_jwt(token)[1]
            for extra in ('iat', 'jti', 'nbf'):
                if extra in decrypted_payload:
                    decrypted_payload.pop(extra)
            assert decrypted_payload == payload


# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-05-01 11:32
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openid_connect_op', '0006_auto_20180103_0922'),
    ]

    operations = [
        migrations.AlterField(
            model_name='openidclient',
            name='sub_hash',
            field=models.CharField(blank=True, max_length=256, null=True, verbose_name='If set, &lt;&lt;sub&gt;&gt; values (that is, username) will be concatenated with this value and sha256'),
        ),
    ]



from django.contrib import admin
from modeltranslation.admin import TranslationAdmin, TranslationStackedInline

from openid_connect_op.models import OpenIDClient, OpenIDAgreement, OpenIDUserAgreement


class OpenIDAgreementAdmin(TranslationStackedInline):
    model = OpenIDAgreement


class OpenIDClientAdmin(TranslationAdmin):
    inlines = (OpenIDAgreementAdmin, )

class OpenIDUserAgreementAdmin(admin.ModelAdmin):
    list_display = ('agreement__client', 'user', 'agreement_obligatory', 'agreement_text')
    list_filter = ('agreement__client__client_name', )
    search_fields = ('user__username', 'user__last_name')
    readonly_fields = ('agreement', 'user', 'agreed_on', 'agreed_by_user')

    def agreement__client(self, instance):
        return instance.agreement.client

    def user(self, instance):
        return instance.user.get_full_name()

    def agreement_obligatory(self, instance):
        return instance.agreement.obligatory

    def agreement_text(self, instance):
        return instance.agreement.text[:50]

admin.site.register(OpenIDClient, OpenIDClientAdmin)
admin.site.register(OpenIDUserAgreement, OpenIDUserAgreementAdmin)


# Generated by Django 2.1.5 on 2019-01-28 17:49
from django.db import migrations
import jsonfield.fields



class Migration(migrations.Migration):

    dependencies = [
        ('openid_connect_op', '0008_auto_20180628_2215'),
    ]

    operations = [
        migrations.AlterField(
            model_name='openidagreement',
            name='allowed_scopes',
            field=jsonfield.fields.JSONField(default=[], verbose_name='List of allowed scopes for this agreement.'),
        ),
    ]



# Generated by Django 2.1.5 on 2019-01-28 17:51

from django.db import migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('openid_connect_op', '0011_auto_20190128_1850'),
    ]

    operations = [
        migrations.AlterField(
            model_name='openidclient',
            name='client_registration_data',
            field=jsonfield.fields.JSONField(default=dict),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-01-03 15:08
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('openid_connect_op', '0004_openidclient_jwks'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='openidkey',
            name='client',
        ),
        migrations.DeleteModel(
            name='OpenIDKey',
        ),
    ]



from django.http import HttpResponseForbidden, JsonResponse
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.csrf import csrf_exempt
from ratelimit.mixins import RatelimitMixin

from openid_connect_op.decorators import access_token_required
from openid_connect_op.models import OpenIDToken
from openid_connect_op.views import OAuthRequestMixin
from openid_connect_op.views.dynamic_registration_view import DynamicClientRegistrationView

try:
    import secrets
except ImportError:
    import openid_connect_op.utils.secrets_backport as secrets


class ClientConfigurationView(RatelimitMixin, OAuthRequestMixin, View):
    ratelimit_key = 'ip'
    ratelimit_rate = '10/m'
    ratelimit_block = True
    ratelimit_method = 'ALL'

    @method_decorator(csrf_exempt)
    @method_decorator(access_token_required())
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, client_id=None):
        if request.openid_access_token and \
                        request.openid_access_token.token_type != OpenIDToken.TOKEN_TYPE_CLIENT_CONFIGURATION_TOKEN:
            return HttpResponseForbidden('Can not use this token to access client configuration view')
        if request.openid_access_token.client.client_id != client_id:
            return HttpResponseForbidden('This token has not been issued for the client identified by client_id %s' % client_id)

        client = request.openid_access_token.client
        # note: we are deviating from the spec that says that
        # "the Client Read Request itself SHOULD NOT cause changes to the Client's registered Metadata values"
        # The reason is that the client secret is never stored in plain in this implementation, only as a hash
        # and it must be returned - so we need to generate a new value
        new_client_secret = secrets.token_urlsafe(32)
        client.set_client_secret(new_client_secret)
        client.save()

        resp = DynamicClientRegistrationView.make_registration_response(request, client, new_client_secret)
        return JsonResponse(resp)
